import asyncio
from asyncio import Semaphore
from aiohttp import ClientSession, ClientTimeout
from concurrent.futures import ProcessPoolExecutor
from typing import List, Iterable, Dict, Callable, Optional, TypeVar

T = TypeVar('T')


class Crawler:
    def __init__(self,
                 concurrent_requests: int = 500,
                 num_parsing_processes: int = 1,
                 request_headers: Optional[Dict[str, str]] = None,
                 request_timeout: Optional[float] = None):
        self.concurrent_requests = concurrent_requests
        self.num_parsing_processes = num_parsing_processes
        self.request_headers = request_headers
        self.request_timeout = request_timeout

    async def _fetch_and_parse(self,
                               sem: Semaphore,
                               pool: ProcessPoolExecutor,
                               sess: ClientSession,
                               url: str,
                               parse_fn: Optional[Callable[[str], T]] = None
                               ) -> str:
        async with sem, sess.get(url) as resp:
            if parse_fn is None:
                return await resp.text()

            # Run `parse_fn` in subprocess from process-pool for parallelism.
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(
                pool, parse_fn, await resp.text())

    async def _crawl_and_reduce(
            self,
            urls: Iterable[str],
            parse_fn: Optional[Callable[[str], T]] = None,
            callback_fn: Optional[Callable[[T], None]] = None
            ):
        # Create a semaphore to limit the number of concurrent requests, a
        # process-pool executor to run `parse_fn` in parallel and a http client
        # session for asynchronous HTTP requests.
        sem = Semaphore(self.concurrent_requests)
        pool = ProcessPoolExecutor(max_workers=self.num_parsing_processes)
        sess = ClientSession(headers=self.request_headers,
                             timeout=ClientTimeout(total=self.request_timeout))

        # Create fetch tasks and add callback functions.
        futures = [asyncio.ensure_future(
                        self._fetch_and_parse(sem, pool, sess, url, parse_fn))
                   for url in urls]
        if callback_fn is not None:
            for f in futures:
                f.add_done_callback(
                    lambda f: callback_fn(f.exception() or f.result()))

        # Wait for the tasks to complete and close the http client session and
        # process-pool executor
        await asyncio.wait(futures)
        await sess.close()
        pool.shutdown(wait=True)

    def reduce_to_array(self,
                        urls: Iterable[str],
                        parse_fn: Optional[Callable[[str], T]] = None,
                        update_fn: Optional[Callable[[], None]] = None
                        ) -> List[T]:
        # A callback function to reduce collected data to the array.
        def callback_fn(data: T):
            if update_fn is not None:
                update_fn()

            if not isinstance(data, Exception):
                results.append(data)

        results = []
        asyncio.run(self._crawl_and_reduce(urls, parse_fn, callback_fn))

        return results

    def reduce_to_file(self,
                       urls: Iterable[str],
                       filename: str,
                       parse_fn: Optional[Callable[[str], T]] = None,
                       update_fn: Optional[Callable[[], None]] = None) -> int:
        with open(filename, 'w') as fp:
            # A callback function to reduce collected data to the output file.
            def callback_fn(data: T):
                if update_fn is not None:
                    update_fn()

                if not isinstance(data, Exception):
                    # Increase the counter which indicates the number of actual
                    # reduced items.
                    nonlocal written
                    written += 1

                    fp.write(str(data) + '\n')

            written = 0
            asyncio.run(self._crawl_and_reduce(urls, parse_fn, callback_fn))

        return written



from datetime import datetime, timedelta
from typing import List


def drange(start: str, end: str, step: int = 1) -> List[str]:
    start_date = datetime.strptime(start, '%Y%m%d')
    end_date = datetime.strptime(end, '%Y%m%d')

    iters = (end_date - start_date).days // step
    return [(start_date + timedelta(days=d * step)).strftime('%Y%m%d')
            for d in range(iters + 1)]


def korean_character_ratio(text: str, ignore_whitespace: bool = True) -> float:
    if ignore_whitespace:
        text = ''.join(text.split())

    korean_characters = len([
        c for c in text if ord('가') <= ord(c) <= ord('힣')])
    return korean_characters / len(text)


def is_normal_character(c: str) -> bool:
    return (ord('0') <= ord(c) <= ord('9')
            or ord('a') <= ord(c) <= ord('z')
            or ord('A') <= ord(c) <= ord('Z')
            or ord('가') <= ord(c) <= ord('힣'))



from setuptools import dist, setup, find_packages

# `Cython` is used when installing `kss` library.
dist.Distribution().fetch_build_eggs(['Cython'])

setup(
    name='canrevan',
    version='3.0.0',

    author='Jungwoo Park',
    author_email='affjljoo3581@gmail.com',

    description='Simple Naver News Crawler',
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type='text/markdown',

    keywords=['naver', 'news', 'dataset', 'nlp'],
    url='https://github.com/affjljoo3581/canrevan',
    license='Apache-2.0',

    package_dir={'': 'src'},
    packages=find_packages('src'),
    python_requires='>=3.6.0',
    install_requires=[
        'tqdm>=4.46.0',
        'bs4',
        'lxml>=4.5.1',
        'aiohttp',
        'langumo'
    ],

    entry_points={
        'console_scripts': [
            'canrevan = canrevan:_main'
        ]
    },

    classifiers=[
        'Environment :: Console',
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ]
)



import re
import json
import canrevan.utils as utils
from bs4 import BeautifulSoup, SoupStrainer
from typing import List


def extract_article_urls(document: str) -> List[str]:
    document = document[document.find('<ul class="type06_headline">'):]

    # Extract article url containers.
    list1 = document[:document.find('</ul>')]
    list2 = document[document.find('</ul>') + 5:]
    list2 = list2[:list2.find('</ul>')]

    document = list1 + list2

    # Extract all article urls from the containers.
    article_urls = []
    while '<dt>' in document:
        document = document[document.find('<dt>'):]
        container = document[:document.find('</dt>')]

        if not container.strip():
            continue

        article_urls.append(
            re.search(r'<a href="(.*?)"', container).group(1))
        document = document[document.find('</dt>'):]

    return article_urls


def parse_article_content(document: str) -> str:
    strainer = SoupStrainer('div', attrs={'id': 'articleBodyContents'})
    document = BeautifulSoup(document, 'lxml', parse_only=strainer)
    content = document.find('div')

    # Skip invalid articles which do not contain news contents.
    if content is None:
        raise ValueError('there is no any news article content.')

    # Remove unnecessary tags except `<br>` elements for preserving line-break
    # characters.
    for child in content.find_all():
        if child.name != 'br':
            child.clear()

    content = content.get_text(separator='\n').strip()
    content = '\n'.join([line.strip() for line in content.splitlines()
                         if line.strip()])

    # Skip the contents which contain too many non-Korean characters.
    if utils.korean_character_ratio(content) < 0.5:
        raise ValueError('there are too few Korean characters in the content.')

    # Normalize the contents by removing abnormal sentences.
    content = '\n'.join([
        line for line in content.splitlines()
        if utils.is_normal_character(line[0]) and line[-1] == '.'])

    return json.encoder.encode_basestring(content)



import tqdm
import argparse
import canrevan.utils as utils
import canrevan.parsing as parsing
from canrevan.crawling import Crawler
from typing import List


def _main():
    args = _create_argument_parser().parse_args()

    # Create a crawler for collecting article urls and news contents.
    crawler = Crawler(concurrent_requests=args.max_jobs,
                      num_parsing_processes=args.num_cores,
                      request_headers={'user-agent': 'canrevan'},
                      request_timeout=args.timeout)

    # Collect article urls from navigation pages.
    nav_urls = _prepare_nav_urls(args)
    print(f'[*] navigation pages: {len(nav_urls)}')

    with tqdm.tqdm(nav_urls, desc='[*] collect article urls') as tbar:
        article_urls = crawler.reduce_to_array(
            nav_urls,
            parse_fn=parsing.extract_article_urls,
            update_fn=tbar.update)

    # Flatten the grouped urls and remove duplicates from the array.
    article_urls = {url for urls in article_urls for url in urls}
    print(f'[*] total collected articles: {len(article_urls)}')

    # Crawl news articles from the collected article urls and save the content
    # to the output file.
    with tqdm.tqdm(article_urls,
                   desc='[*] crawl news article contents') as tbar:
        total_contents = crawler.reduce_to_file(
            article_urls,
            args.output_path,
            parse_fn=parsing.parse_article_content,
            update_fn=tbar.update)

    print(f'[*] finish crawling {total_contents} news articles to '
          f'[{args.output_path}]')


def _prepare_nav_urls(args: argparse.Namespace) -> List[str]:
    return [f'https://news.naver.com/main/list.nhn?mode=LSD&mid=shm'
            f'&sid1={category}&date={date}&page={page}'
            for category in args.category
            for date in utils.drange(args.start_date,
                                     args.end_date,
                                     args.skip_days)
            for page in range(1, args.max_page + 1)]


def _create_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog='canrevan', description='crawl naver news articles')

    parser.add_argument('--output_path', default='articles.txt',
                        help='output file path')
    parser.add_argument('--category', required=True, nargs='*', type=int,
                        help='list of news article categories')
    parser.add_argument('--start_date', required=True,
                        help='minimum date of news articles')
    parser.add_argument('--end_date', required=True,
                        help='maximum date of news articles')
    parser.add_argument('--skip_days', default=1, type=int,
                        help='number of days to skip from crawling')
    parser.add_argument('--max_page', default=10, type=int,
                        help='maximum number of pages to navigate')
    parser.add_argument('--timeout', default=5, type=float,
                        help='timeout for the whole request')
    parser.add_argument('--max_jobs', default=500, type=int,
                        help='maximum number of concurrent requests')
    parser.add_argument('--num_cores', default=4, type=int,
                        help='number of multi-processing cores for parsing')

    return parser
